"""**rebar** helps with reinforcement. That's why it's called rebar! It's a toolkit that has evolved
as I've worked on RL projects.

Unlike the :mod:`megastep` module which is stable, documented and feature-complete, rebar is an unstable,
undocumented work-in-progress. It's in the megastep repo because megastep itself uses two of rebar's most useful components:
:class:`~rebar.dotdict.dotdict` and :class:`~rebar.arrdict.arrdict`, while the demo uses a whole lot more.
"""